
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.InputStream;

class SchemeScannerImpl implements SchemeScanner {
    private String sourceName;
    private BufferedReader source;
    private int line;
    private int ch;
    private int toBeRead;
    private int token;
    private String value;

    public SchemeScannerImpl(String n, InputStream s) throws java.io.IOException {
	sourceName = n;
	source = new BufferedReader(new InputStreamReader(s));
	line = 1;
	ch = 0;
	toBeRead = 1;
	nextToken();
    }

    private static boolean isInitialIdChar(char ch) {
	return Character.isLetter(ch) || "!$%&*/:<=>?^_~+-".indexOf(ch) >= 0;
    }

    private static boolean isSubsequentIdChar(char ch) {
	return isInitialIdChar(ch) || Character.isDigit(ch) || ".@".indexOf(ch) >= 0;
    }

    public boolean more() throws java.io.IOException {
	return ch >= 0;
    }

    private char getChar() throws java.io.IOException {
	while (ch >= 0 && toBeRead > 0) {
	    ch = source.read();
	    if (ch == '\n') line++;
	    toBeRead--;
	}
	return (char) ch;
    }

    private void nextChar() {
	toBeRead++;
    }

    private void skipSpaces() throws java.io.IOException {
	while (true) {
	    while (more() && Character.isWhitespace(getChar()))
		nextChar();
	    if (getChar() == ';')
		while (more() && getChar() != '\n')
		    nextChar();
	    else
		break;
	}
    }

    private void setToken(int t, String v) {
	token = t;
	value = v;
    }

    private void setToken(int t) {
	setToken(t, "");
    }

    public int getToken() {
	return token;
    }

    public String getValue() {
	return value;
    }

    public String getSourceName() {
	return sourceName;
    }

    public int getLine() {
	return line;
    }

    public void printToken() {
	switch (getToken()) {
	case OPEN:
	    System.out.println("OPEN");
	    break;
	case CLOSE:
	    System.out.println("CLOSE");
	    break;
	case ID:
	    System.out.println("ID " + getValue());
	    break;
	case BOOL:
	    System.out.println("BOOL " + getValue());
	    break;
	case INT:
	    System.out.println("INT " + getValue());
	    break;
	case STRING:
	    System.out.println("STRING " + getValue());
	    break;
	}
    }

    private static boolean isInt(String s) {
	for (int i = 0; i < s.length(); i++)
	    if (!Character.isDigit(s.charAt(i)))
		return false;
	return true;
    }

    public void nextToken() throws java.io.IOException {
	skipSpaces();
	if (more()) {
	    switch (getChar()) {
	    case '#':
		nextChar();
		switch (getChar()) {
		case 't': 
		    nextChar();
		    setToken(BOOL, "t");
		    break;
		case 'f':
		    nextChar();
		    setToken(BOOL, "f");
		    break;
		default:
		    break; // error
		}
		    break;
	    case '(':
		nextChar();
		setToken(OPEN);
		break;
	    case ')':
		nextChar();
		setToken(CLOSE);
		break;
	    case '\'': {
		nextChar();
		final StringBuffer buffer = new StringBuffer();
		while (more() && isSubsequentIdChar(getChar())) {
		    buffer.append(getChar());
		    nextChar();
		}
		if (buffer.length() == 0)
		    throw new SchemeSyntaxError(getSourceName(), getLine(), "void symbol");
		setToken(SYMBOL, buffer.toString());
		break;
	    }
	    case '"': {
		nextChar();
		final StringBuffer buffer = new StringBuffer();
		while (more() && getChar() != '"') {
		    buffer.append(getChar());
		    nextChar();
		}
		if (more() && getChar() == '"')
		    nextChar();
		else
		    throw new SchemeSyntaxError(getSourceName(), getLine(), "open string");
		setToken(STRING, buffer.toString());
		break;
	    }
	    default:
		if (Character.isDigit(getChar())) {
		    StringBuffer buffer = new StringBuffer();
		    buffer.append(getChar());
		    nextChar();
		    while (more() && Character.isDigit(getChar())) {
			buffer.append(getChar());
			nextChar();
		    }
		    setToken(INT, buffer.toString());
		} else if (isInitialIdChar(getChar())) {
		    StringBuffer buffer = new StringBuffer();
		    while (more() && isSubsequentIdChar(getChar())) {
			buffer.append(getChar());
			nextChar();
		    }
		    if (buffer.length() > 1 && buffer.charAt(0) == '-' && isInt(buffer.substring(1)))
			setToken(INT, buffer.toString());
		    else
			setToken(ID, buffer.toString());		
		} else
		    throw new SchemeSyntaxError(getSourceName(), getLine(),
						"unrecognized character `" + String.valueOf(getChar()) + "'");
	    }
	} else
	    setToken(EOF);
    }
}
